/////////////////////////////////////////////////////////////////////////////
// Name:        SVGCanvasAgg.cpp
// Purpose:     Agg render
// Author:      Alex Thuering
// Created:     2005/05/04
// RCS-ID:      $Id: SVGCanvasAgg.cpp,v 1.8 2005/11/17 17:53:28 ntalex Exp $
// Copyright:   (c) 2005 Alex Thuering
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

#include "SVGCanvasAgg.h"
#include "SVGCanvasPathAgg.h"
#ifdef __WXMSW__
#include "SVGCanvasTextAgg.h"
#else
#include "../freetype/SVGCanvasTextFreetype.h"
#endif

#include <wx/log.h>
#include <agg2/agg_rasterizer_scanline_aa.h>
#include <agg2/agg_scanline_p.h>
#include <agg2/agg_scanline_u.h>
#include <agg2/agg_span_interpolator_linear.h>
#include <agg2/agg_span_interpolator_trans.h>
#include <agg2/agg_span_image_filter_rgb.h>
#include <agg2/agg_span_converter.h>
	
wxSVGCanvasAgg::wxSVGCanvasAgg(): m_rbuf(NULL), m_pixf(NULL),
  m_rendererBase(NULL), m_rendererSolid(NULL)
{
}

wxSVGCanvasAgg::~wxSVGCanvasAgg()
{
  delete m_rbuf;
  delete m_pixf;
  delete m_rendererBase;
  delete m_rendererSolid;
}

void wxSVGCanvasAgg::Init(int width, int height)
{
  m_image.Create(width, height);
  m_rbuf = new RenderingBuffer(m_image.GetData(), 
	 m_image.GetWidth(), m_image.GetHeight(), m_image.GetWidth()*3);
  m_pixf = new PixFmt(*m_rbuf);
  m_rendererBase = new RendererBase(*m_pixf);
  m_rendererSolid = new RendererSolid(*m_rendererBase);
}

wxImage wxSVGCanvasAgg::GetImage()
{
  return m_image;
}

void wxSVGCanvasAgg::Clear(wxRGBColor color)
{
  if (!m_rendererBase)
    return;
  m_rendererBase->clear(agg::rgba(
    color.Red()/255, color.Green()/255, color.Blue()/255));
}

wxSVGCanvasPath* wxSVGCanvasAgg::CreateCanvasPath()
{
  return new wxSVGCanvasPathAgg();
}

wxSVGCanvasItem* wxSVGCanvasAgg::CreateItem(wxSVGTextElement* element,
  wxCSSStyleDeclaration* style)
{
#ifdef __WXMSW__
  wxSVGCanvasText* canvasText = new wxSVGCanvasTextAgg(this);
#else
  wxSVGCanvasText* canvasText = new wxSVGCanvasTextFreetype(this);
#endif
  if (style == NULL)
	style = (wxCSSStyleDeclaration*) &element->GetStyle();
  canvasText->Init(*element, *style);
  return canvasText;
}

void wxSVGCanvasAgg::DrawItem(wxSVGCanvasItem& item,
  wxSVGMatrix& matrix, wxCSSStyleDeclaration& style)
{
  switch (item.GetType())
  {
	case wxSVGCanvasItem::wxSVG_CANVAS_ITEM_PATH:
	  DrawCanvasPath((wxSVGCanvasPathAgg&) item, matrix, style);
	  break;
    case wxSVGCanvasItem::wxSVG_CANVAS_ITEM_TEXT:
      DrawCanvasText((wxSVGCanvasText&) item, matrix, style);
	  break;
	case wxSVGCanvasItem::wxSVG_CANVAS_ITEM_IMAGE:
	  DrawCanvasImage((wxSVGCanvasImage&) item, matrix, style);
	  break;
  }
}

void wxSVGCanvasAgg::DrawCanvasPath(wxSVGCanvasPathAgg& canvasPath,
 wxSVGMatrix& matrix, wxCSSStyleDeclaration& style)
{
  if (!m_rendererBase)
    return;
  rasterizer_scanline_aa<> ras;
  scanline_p8 sl;
  
  const rect& cb = m_rendererBase->clip_box();
  ras.clip_box(cb.x1, cb.y1, cb.x2+1, cb.y2+1);
	  
  canvasPath.m_transform = trans_affine(matrix.GetA(), matrix.GetB(),
	matrix.GetC(), matrix.GetD(), matrix.GetE(), matrix.GetF());
	  
  if (style.GetFill().Ok() &&
	  style.GetFill().GetPaintType() != wxSVG_PAINTTYPE_URI &&
	  style.GetFill().GetPaintType() != wxSVG_PAINTTYPE_URI_NONE)
  {
	ras.reset();
	ras.filling_rule(
	  style.GetFillRule() == wxCSS_VALUE_EVENODD ? fill_even_odd : fill_non_zero);
	if (fabs(canvasPath.m_curved_trans_contour.width()) < 0.0001)
	  ras.add_path(canvasPath.m_curved_trans);
	else
	{
	  canvasPath.m_curved_trans_contour.miter_limit(style.GetStrokeMiterlimit());
	  ras.add_path(canvasPath.m_curved_trans_contour);
	}
	
	const wxRGBColor& c = style.GetFill().GetRGBColor();
	rgba8 color(c.Red(), c.Green(), c.Blue());
	color.opacity(style.GetFillOpacity());
	m_rendererSolid->color(color);
	agg::render_scanlines(ras, sl, *m_rendererSolid);
  }

  if (style.GetStroke().Ok() && style.GetStrokeWidth()>0 &&
	  style.GetStroke().GetPaintType() != wxSVG_PAINTTYPE_URI &&
	  style.GetStroke().GetPaintType() != wxSVG_PAINTTYPE_URI_NONE)
  {
	canvasPath.m_curved_stroked.width(style.GetStrokeWidth());
	if (style.GetStrokeLinejoin() == wxCSS_VALUE_MITER)
	  canvasPath.m_curved_stroked.line_join(miter_join);
	else if (style.GetStrokeLinejoin() == wxCSS_VALUE_ROUND)
	  canvasPath.m_curved_stroked.line_join(round_join);
	else if (style.GetStrokeLinejoin() == wxCSS_VALUE_BEVEL)
	  canvasPath.m_curved_stroked.line_join(bevel_join);
	if (style.GetStrokeLinecap() == wxCSS_VALUE_BUTT)
	  canvasPath.m_curved_stroked.line_cap(butt_cap);
	else if (style.GetStrokeLinecap() == wxCSS_VALUE_ROUND)
	  canvasPath.m_curved_stroked.line_cap(round_cap);
	else if (style.GetStrokeLinecap() == wxCSS_VALUE_SQUARE)
	  canvasPath.m_curved_stroked.line_cap(square_cap);
	canvasPath.m_curved_stroked.miter_limit(style.GetStrokeMiterlimit());
	ras.reset();
	ras.filling_rule(fill_non_zero);
	ras.add_path(canvasPath.m_curved_stroked_trans);
	const wxRGBColor& c = style.GetStroke().GetRGBColor();
	rgba8 color(c.Red(), c.Green(), c.Blue());
	color.opacity(style.GetStrokeOpacity());
	m_rendererSolid->color(color);
	agg::render_scanlines(ras, sl, *m_rendererSolid);
  }
}

class SpanConvAlpha
{
  public:
	SpanConvAlpha::SpanConvAlpha(double alpha)
	{
	  m_alpha = alpha;
	}
	
	void convert(rgba8* colors, int x, int y, unsigned len) const
	{
	  while (len)
	  {
		colors->a = (unsigned char) (m_alpha*colors->a);
		colors++;
		len--;
	  }
	}
  
  protected:
	double m_alpha;
};

void wxSVGCanvasAgg::DrawCanvasImage(wxSVGCanvasImage& canvasImage,
  wxSVGMatrix& matrix, wxCSSStyleDeclaration& style)
{
  if (!canvasImage.m_image.Ok() ||
      canvasImage.m_image.GetWidth() == 0 ||
      canvasImage.m_image.GetHeight() == 0)
    return;
  
  // source_buffer
  rendering_buffer source_buffer;
  source_buffer.attach(canvasImage.m_image.GetData(),
	canvasImage.m_image.GetWidth(), canvasImage.m_image.GetHeight(),
	canvasImage.m_image.GetWidth()*3);
	
  //transform
  wxSVGMatrix transMatrix = matrix.Translate(canvasImage.m_x, canvasImage.m_y);
  transMatrix = transMatrix.ScaleNonUniform(
    canvasImage.m_width/canvasImage.m_image.GetWidth(),
    canvasImage.m_height/canvasImage.m_image.GetHeight());
  trans_affine transform(transMatrix.GetA(), transMatrix.GetB(),
	transMatrix.GetC(), transMatrix.GetD(), transMatrix.GetE(), transMatrix.GetF());
  double x1 = 0;
  double y1 = 0;
  double x2 = canvasImage.m_image.GetWidth();
  double y2 = canvasImage.m_image.GetHeight();
  transform.transform(&x1, &y1);
  transform.transform(&x2, &y2);
  transform.invert();
  
  scanline_u8 sl; // better as scanline_p8 because the number of spans
  rasterizer_scanline_aa<> ras;
  const rect& cb = m_rendererBase->clip_box();
  ras.clip_box(cb.x1, cb.y1, cb.x2+1, cb.y2+1);
  
  path_storage imgRect;
  imgRect.move_to(x1, y1);
  imgRect.line_to(x1, y2);
  imgRect.line_to(x2, y2);
  imgRect.line_to(x2, y1);
  imgRect.close_polygon();
  ras.add_path(imgRect);
  
  double opacity = style.GetOpacity();
  if (opacity == 1)
  {
	typedef span_interpolator_linear<> InterpolatorType;
	typedef span_image_filter_rgb_bilinear<rgba8, order_rgb, InterpolatorType> SpanGenType;
	typedef renderer_scanline_aa<RendererBase, SpanGenType> RendererType;
  
	InterpolatorType interpolator(transform);
	span_allocator<rgba8> sa;
	SpanGenType sg(sa, source_buffer, rgba(1, 1, 1, 0), interpolator);
	RendererType ri(*m_rendererBase, sg);
	render_scanlines(ras, sl, ri);
  }
  else
  {
	typedef span_interpolator_linear<> InterpolatorType;
	typedef span_image_filter_rgb_bilinear<rgba8, order_rgb, InterpolatorType> SpanGenType;
	typedef span_converter<SpanGenType, SpanConvAlpha> SpanConv;
	typedef renderer_scanline_aa<RendererBase, SpanConv> RendererType;
  
	InterpolatorType interpolator(transform);
	span_allocator<rgba8> sa;
	SpanGenType sg(sa, source_buffer, rgba(1, 1, 1, 0), interpolator);
	SpanConvAlpha spanConvAlpha(opacity);
	SpanConv sc(sg, spanConvAlpha);
	RendererType ri(*m_rendererBase, sc);
	render_scanlines(ras, sl, ri);
  }
}
